__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import shlex
import threading
import time

import tunel.utils as utils
from tunel.launcher.base import Launcher
from tunel.logger import logger


class Singularity(Launcher):
    def run(self, *args, **kwargs):
        """
        Run handles some command to singularity (e.g., run or exec)
        """
        command = "%s %s singularity %s" % (self.path, self.environ, " ".join(args[0]))

        # If the user wants a shell, give to them!
        if "shell" in args[0]:
            self.ssh.shell(command, interactive=True)
        else:
            res = self.ssh.execute(command)
            self.ssh.print_output(res)

    def run_app(self, app):
        """
        A Singularity app means running a container directly with some arguments, etc.
        """
        # Make sure we set the username to the ssh
        self.username

        # Add any paths from the config
        paths = self.settings.get("paths", [])

        # Prepare dictionary with content to render into recipes
        render = self.prepare_render(app, paths)

        # Clean up previous sockets
        self.ssh.execute(["rm", "-rf", "%s/*.sock" % render["scriptdir"]])

        # Load the app template
        template = app.load_template()
        result = template.render(**render)

        # Write script to temporary file
        tmpfile = utils.get_tmpfile()
        utils.write_file(tmpfile, result)

        # Copy over to server
        remote_script = os.path.join(self.remote_assets_dir, app.name, app.script)
        self.ssh.scp_to(tmpfile, remote_script)

        # Instead of a Singularity command, we run the script
        logger.c.print()
        logger.c.print("== INSTRUCTIONS WILL BE PRINTED with a delay ==")
        command = "%s %s %s %s" % (
            self.path,
            self.environ,
            self.ssh.settings.shell,
            remote_script,
        )
        self.print_tunnel_instructions(app, render["socket"])
        self.ssh.execute(command, stream=True)

    def print_tunnel_instructions(self, app, socket):
        """
        Start a separate thread to print connection details.
        """
        thread = threading.Thread(
            target=post_commands,
            name="Logger",
            args=[self.ssh, app, socket],
        )
        thread.start()


def post_commands(ssh, app, socket):
    time.sleep(10)
    ssh.tunnel_login_node(socket=socket, app=app)
    if app.post_command:
        logger.info("Found post command %s" % app.post_command)
        post = app.post_command.replace("$socket_dir", os.path.dirname(socket))
        ssh.execute(shlex.split(post), stream=True)
    time.sleep(30)
    ssh.tunnel_login_node(socket=socket, app=app)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Get info for an app

import json

import tunel.apps


def main(args, parser, extra, subparser):
    app = tunel.apps.get_app(args.app, extra)
    if args.json:
        print(json.dumps(app.config, indent=4))
        return

    app.print_pretty()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

__version__ = "0.0.16"
AUTHOR = "Vanessa Sochat"
NAME = "tunel"
PACKAGE_URL = "https://github.com/tunel-apps/tunel"
KEYWORDS = "interface, ui, developer, HPC, containers"
DESCRIPTION = "Developer interfaces for HPC via tunneling"
LICENSE = "LICENSE"

################################################################################
# Global requirements

# Since we assume wanting Singularity and lmod, we require spython and Jinja2

INSTALL_REQUIRES = (
    ("jsonschema", {"min_version": None}),
    ("pyaml", {"min_version": None}),
    ("jinja2", {"min_version": None}),
    ("rich", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (versions that include database)

INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import json
import os

import tunel.api
from tunel.logger import logger


def api_get(args, parser, extra, subparser):
    if not args.socket:
        logger.exit("Socket for tunel must be provided with --socket")
    if not os.path.exists(args.socket):
        logger.exit("Socket for tunel %s does not exist." % args.socket)
    api = tunel.api.ApiConnection(args.socket)
    res = api.get(args.path)
    if args.json:
        print(json.dumps(res.json(), indent=4))
    else:
        print(res.text)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import shlex
import threading
import time

import tunel.utils
from tunel.launcher.base import Launcher
from tunel.logger import logger

here = os.path.dirname(os.path.abspath(__file__))


class Slurm(Launcher):
    """
    A slurm launcher interacts with slurm
    """

    def __init__(self, server, **kwargs):
        super().__init__(server, **kwargs)
        self._inventory = {}

    @property
    def modules_file(self):
        return os.path.join(self.assets_dir, "modules.txt")

    @property
    def nodes_file(self):
        return os.path.join(self.assets_dir, "sinfo.txt")

    def update_inventory(self):
        if self._inventory:
            return
        self._update_inventory_modules()
        self._update_inventory_nodes()

    def _update_inventory_nodes(self):
        """
        Try to derive attributes for nodes
        """
        if not os.path.exists(self.nodes_file):
            res = self.ssh.execute("scontrol show config")
            tunel.utils.write_file(self.nodes_file, res["message"])
        if os.path.exists(self.nodes_file) and "nodes" not in self._inventory:
            self._inventory["nodes"] = tunel.utils.read_config_file(self.nodes_file)

    def _update_inventory_modules(self):
        """
        Try to derive list of installed modules on cluster.
        This has been tested for LMOD
        """
        # if we don't have modules, write there
        if not os.path.exists(self.modules_file):
            res = self.scp_and_run("list_modules.sh")
            if res["return_code"] == 0:
                res = [x for x in res["message"].split("\n") if "RESULT:" in x]

                # This is the output file with list of modules
                if res:
                    res = res[0].split(":")[-1]

                    # scp get is copying FROM the server to assets here
                    self.scp_get(res, self.modules_file)

        # When we get here, only will exist if command was successful
        if os.path.exists(self.modules_file) and "modules" not in self._inventory:
            self._inventory["modules"] = tunel.utils.read_lines(self.modules_file)

    def _prepare_render(self, app, paths):
        """
        Given an app, prepare default variables (and custom args) to render
        """
        render = {}
        # TODO needs should not be hard coded for slurm
        if app.needs:
            render["modules"] = self.get_modules(app.needs.get("modules"))
        return render

    def run_app(self, app):
        """
        Given an app designated for slurm, run it!
        """
        self.update_inventory()

        # Add any paths from the config
        paths = self.settings.get("paths", [])

        # Prepare dictionary with content to render into recipes
        render = self.prepare_render(app, paths)
        remote_script = render["script"]

        # Clean up previous sockets
        self.ssh.execute(["rm", "-rf", "%s/*.sock" % render["scriptdir"]])

        # Load the app template
        template = app.load_template()
        result = template.render(**render)

        # Write script to temporary file
        tmpfile = self.write_temporary_script(result)

        # Copy over to server
        self.ssh.scp_to(tmpfile, render["script"])
        os.remove(tmpfile)

        # Assemble the command
        command = [
            "sbatch",
            "--job-name=%s" % app.job_name,
            "--output=%s" % render["log_output"],
            "--error=%s" % render["log_error"],
            remote_script,
        ]

        # Launch with command
        if not self.previous_job_exists(app.job_name):
            self.run(
                command,
                app.job_name,
                logs_prefix=remote_script,
                app=app,
                socket=render["socket"],
            )

    def stop_app(self, app):
        """
        Wrapper to stop a single app.
        """
        return self.stop([app.name])

    def stop(self, names):
        """
        Stop one or more named jobs
        """
        for name in names:
            logger.purple("Killing %s slurm job on %s" % (name, self.ssh.server))
            self.ssh.execute_or_fail(
                "squeue --name %s --user %s" % (name, self.username)
                + " -o '%A' -h | xargs --no-run-if-empty /usr/bin/scancel"
            )
            logger.purple("Killing listeners on %s" % self.ssh.server)
            self.ssh.execute_or_fail(
                "lsof -i :%s -t | xargs --no-run-if-empty kill" % self.ssh.remote_port
            )

    def previous_job_exists(self, name):
        """
        Check to see if a previous job was submit/exists.
        """
        logger.purple("Checking for previous run of %s" % name)
        previous = self.ssh.execute_or_fail(
            "squeue --name %s --user %s" % (name, self.username) + " -o %R -h"
        )
        if previous == "":
            logger.purple("No existing %s jobs found, continuing." % name)
        else:
            logger.error("Found existing job for %s!" % name)
            logger.exit(
                "Please run 'tunel stop-slurm %s %s' before proceeding."
                % (self.ssh.server, name)
            )

    def get_machine(self, name, max_attempts=None):
        """
        Given the name of a job, wait for the job to start and return the machine
        """
        max_attempts = max_attempts or self.settings.get("max_attempts")
        timeout = 2
        attempt = 1
        machine = ""
        allocated = False

        logger.purple("Waiting for job to start, using exponential backoff")

        # Continue waiting until machine is allocated
        while not allocated:
            with logger.c.status("Waiting...", spinner=self.ssh.settings.tunel_spinner):
                machine = self.ssh.execute_or_fail(
                    "squeue --name %s --user %s" % (name, self.username) + " -o %N -h"
                )
                if machine != "":
                    logger.purple(
                        "Attempt %s: resources allocated to %s!" % (attempt, machine)
                    )
                    allocated = True
                    break

                logger.purple(
                    "Attempt %s: not ready yet... retrying in %s..."
                    % (attempt, timeout)
                )
                time.sleep(timeout)
                attempt += 1
                if max_attempts and attempt >= max_attempts:
                    logger.exit(
                        f"Max attempts {max_attempts} reached, stopping trying."
                    )
                timeout = timeout * 2

        return machine

    def show_logs_instruction(self, logs_prefix):
        """
        print to the screen how to see currently running logs
        """
        logger.purple("View logs in separate terminal")
        logger.info("ssh %s cat %s.out" % (self.ssh.server, logs_prefix))
        logger.info("ssh %s cat %s.err" % (self.ssh.server, logs_prefix))
        print()

    def print_updated_logs(self, logs_prefix, app, socket):
        """
        Start a separate thread that regularly checks and prints logs (when there is an updated line)
        """
        logs_thread = threading.Thread(
            target=post_commands,
            name="Logger",
            args=[self.ssh, app, logs_prefix, socket],
        )
        logs_thread.start()

    def run(self, cmd, job_name=None, logs_prefix=None, app=None, socket=None):
        """
        Run a command for slurm, typically sbatch (and eventually with supporting args)
        """
        self.update_inventory()

        # Generate a name for the job, if not supplied
        if not job_name:
            job_name = tunel.utils.namer.generate()

        # If no command, get interactive node
        if not cmd:
            logger.info("No command supplied, will init interactive session!")
            self.ssh.shell("srun --pty bash", interactive=True)

        else:
            if "sbatch" not in cmd:
                cmd = ["sbatch"] + cmd
            res = self.ssh.execute(" ".join(cmd))
            self.ssh.print_output(res)

            # A successful submission should:
            if res["return_code"] == 0:

                # 1. Show the user how to quickly get logs (if logs_prefix provided)
                if logs_prefix:
                    self.show_logs_instruction(logs_prefix)

                machine = self.get_machine(job_name)
                logger.info("%s is running on %s!" % (job_name, machine))

                # Setup port forwarding
                time.sleep(10)
                logger.c.print("== Connecting to %s ==" % job_name)
                logger.c.print("== Instructions ==")
                logger.c.print(
                    "1. Password, output, and error will print to this - [bold]make sure application is ready before interaction."
                )
                logger.c.print(
                    "2. Browser: http://%s:%s/ -> http://localhost:%s/..."
                    % (machine, self.ssh.remote_port, self.ssh.local_port)
                )
                logger.c.print(
                    "3. To end session: tunel stop-slurm %s %s"
                    % (self.ssh.server, job_name)
                )
                # Create another process to check logs?
                if logs_prefix:
                    self.print_updated_logs(logs_prefix, app, socket=socket)
                self.ssh.tunnel(machine, socket=socket, app=app)


def post_commands(ssh, app, logs_prefix, socket):
    """
    Post commands to show logs and any commands->post defined by the app
    """
    output_command = "ssh %s tail -3 %s.out" % (ssh.server, logs_prefix)
    error_command = "ssh %s tail -3 %s.err" % (ssh.server, logs_prefix)

    last_out = ""
    last_err = ""
    post_command = False
    while True:
        time.sleep(5)
        new_out = ssh.execute_or_fail(output_command, quiet=True)
        new_err = ssh.execute_or_fail(error_command, quiet=True)
        panels = {}

        # Show a post command (once), if defined
        if not post_command and app.post_command:
            logger.info("Found post command %s" % app.post_command)
            post = app.post_command.replace("$socket_dir", os.path.dirname(socket))
            ssh.execute(shlex.split(post), stream=True)
            post_command = True

        if new_out and new_out != last_out:
            panels["cyan"] = output_command + "\n" + new_out
            last_out = new_out
        if new_err and new_err != last_err:
            panels["magenta"] = error_command + "\n" + new_err
            last_err = new_err
        if panels:
            logger.panel_group(panels)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
from datetime import datetime

import jsonschema
import yaml

import tunel.defaults as defaults
import tunel.schemas
from tunel.logger import logger


class SettingsBase:
    def __init__(self):
        """
        Create a new settings object not requiring a settings file.
        """
        self.settings_file = None

    def __str__(self):
        return "[tunel-settings]"

    def __repr__(self):
        return self.__str__()

    def validate(self):
        """
        Validate the loaded settings with jsonschema
        """
        jsonschema.validate(instance=self._settings, schema=tunel.schemas.settings)

    def edit(self):
        """
        Interactively edit a config file.
        """
        if not self.settings_file or not os.path.exists(self.settings_file):
            logger.exit("Settings file not found.")

        # Make sure editor exists first!
        editor = tunel.utils.which(self.config_editor)
        if editor["return_code"] != 0:
            logger.exit(
                "Editor '%s' not found! Update settings.yml to an editor you have."
                % self.config_editor
            )
        tunel.utils.run_command([self.config_editor, self.settings_file], stream=True)

    def get_settings_file(self, settings_file=None):
        """
        Get the preferred used settings file.
        """
        # Only consider user settings if the file exists!
        user_settings = None
        if os.path.exists(defaults.user_settings_file):
            user_settings = defaults.user_settings_file

        # First preference to command line, then user settings, then default
        return settings_file or user_settings or defaults.default_settings_file

    def load(self, settings_file=None):
        """
        Load the settings file into the settings object
        """
        # Get the preferred settings flie
        self.settings_file = self.get_settings_file(settings_file)

        # Exit quickly if the settings file does not exist
        if not os.path.exists(self.settings_file):
            logger.exit("%s does not exist." % self.settings_file)

        # Store the original settings for update as we go
        self._settings = tunel.utils.read_yaml(self.settings_file)

    def get(self, key, default=None):
        """
        Get a settings value, doing appropriate substitution and expansion.
        """
        value = self._settings.get(key, default)
        value = self._substitutions(value)

        # If we allow environment substitution, do it
        if key in defaults.allowed_envars and value:
            if isinstance(value, list):
                value = [os.path.expandvars(v) for v in value]
            else:
                value = os.path.expanduser(os.path.expandvars(value))
        return value

    def __getattr__(self, key):
        """
        A direct get of an attribute, but default to None if doesn't exist
        """
        return self.get(key)

    def change_validate(self, key, value):
        """
        A courtesy function to validate a new config addition.
        """
        # Don't allow the user to add a setting not known
        try:
            self.validate()
        except jsonschema.exceptions.ValidationError as error:
            logger.exit(
                "%s:%s cannot be added to config: %s" % (key, value, error.message)
            )

    def _substitutions(self, value):
        """
        Given a value, make substitutions
        """
        if isinstance(value, (bool, int)) or not value:
            return value

        # Currently dicts only support boolean or null so we return as is
        elif isinstance(value, dict):
            return value

        for rep, repvalue in defaults.reps.items():
            if isinstance(value, list):
                value = [x.replace(rep, repvalue) for x in value]
            else:
                value = value.replace(rep, repvalue)

        return value

    def delete(self, key):
        if key in self._settings:
            del self._settings[key]

    def save(self, filename=None):
        """
        Save settings, but do not change order of anything.
        """
        filename = filename or self.settings_file
        if not filename:
            logger.exit("A filename is required to save to.")

        with open(filename, "w") as fd:
            yaml.dump(self._settings, fd)

    def __iter__(self):
        for key, value in self.__dict__.items():
            yield key, value


class Settings(SettingsBase):
    """
    The settings class is a wrapper for easily parsing a settings.yml file.

    We parse into a query-able class. It also gives us control to update settings,
    meaning we change the values and then write them to file. It's basically
    a dictionary-like class with extra functions.
    """

    def __init__(self, settings_file, validate=True):
        """
        Create a new settings object, which requires a settings file to load
        """
        self.load(settings_file)
        if validate:
            self.validate()

        # Set an updated time, in case it's written back to file
        self._settings["updated_at"] = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os

import tunel.ssh
import tunel.utils as utils

here = os.path.dirname(os.path.abspath(__file__))


class Launcher:
    """
    A launcher is a base for executing commands.
    """

    def __init__(self, server, **kwargs):
        """
        The Launcher is not responsible for managing or closing the connection.
        """
        self.ssh = tunel.ssh.Tunnel(server, **kwargs)

        # A launcher can control assets locally or remotely
        self._remote_work = None
        self._remote_home = None
        self._remote_user = None
        self._home = None

        # Add launcher specific settings, if they exist.
        self.settings = {}
        if self.slug in self.ssh.settings.launchers:
            self.settings = self.ssh.settings.launchers[self.slug]

    @property
    def assets(self):
        return os.path.join(here, "scripts", self.slug)

    @property
    def slug(self):
        return str(self).lower()

    def __str__(self):
        return str(self.__class__.__name__)

    def run(self, *args, **kwargs):
        raise NotImplementedError

    def run_app(self, *args, **kwargs):
        raise NotImplementedError

    def stop(self, *args, **kwargs):
        raise NotImplementedError

    def stop_app(self, *args, **kwargs):
        raise NotImplementedError

    def write_temporary_script(self, content):
        """
        Write a temporary script to file
        """
        tmpfile = utils.get_tmpfile()
        utils.write_file(tmpfile, content)
        return tmpfile

    def get_modules(self, needed_modules):
        """
        Given a list of needed modules, return for an app to load
        """
        if not needed_modules:
            return
        modules = self._inventory.get("modules")
        found_modules = []
        for needed_module in needed_modules:
            matches = [x for x in modules if needed_module in x]
            if matches:
                # Choose the last (likely latest version)
                found_modules += matches[-1]
        return found_modules

    def prepare_render(self, app, paths):
        """
        Given an app, prepare default variables (and custom args) to render
        """
        render = {"args": app.args}
        slug = app.name.replace(os.sep, "-").replace("/", "-")
        if app.needs:
            paths += app.needs.get("paths", [])
        render["jobslug"] = slug
        render["jobname"] = app.name
        render["port"] = self.ssh.remote_port
        render["scriptdir"] = os.path.join(self.remote_assets_dir, app.name)
        render["paths"] = paths
        render["socket"] = os.path.join(render["scriptdir"], "%s.sock" % slug)

        # Remote script, output and error files
        render["script"] = os.path.join(self.remote_assets_dir, app.name, app.script)
        render["script_basename"] = app.script
        render["log_output"] = render["script"] + ".out"
        render["log_error"] = render["script"] + ".err"
        render["log_file"] = render["script"] + ".log"

        # This is second priority to args.workdir
        if self.remote_work:
            render["workdir"] = self.remote_work
        if "workdir" in render["args"]:
            render["workdir"] = render["args"]

        # Does the subclass have customizations?
        if hasattr(self, "_prepare_render"):
            render.update(self._prepare_render(app, paths))
        return render

    @property
    def path(self):
        """
        Get additions to the path
        """
        paths = self.settings.get("paths", "")
        if paths:
            paths = "PATH=%s:$PATH" % (":".join(paths))
        return paths

    @property
    def environ(self):
        """
        Get envars
        """
        envars = self.settings.get("environment", "")
        if envars:
            envars = " ".join(envars)
        return envars

    @property
    def assets_dir(self):
        return os.path.join(self.home, "tunel", self.slug)

    @property
    def remote_assets_dir(self):
        return os.path.join(self.remote_home, "tunel", self.slug)

    @property
    def home(self):
        """
        Get (or create) a local home
        """
        if not self._home:
            self._home = self.ssh.settings.tunel_home
            if not os.path.exists(self.assets_dir):
                os.makedirs(self.assets_dir)
        return self._home

    @property
    def username(self):
        """
        Get the username
        """
        if not self._remote_user:
            self._remote_user = self.ssh.execute_or_fail("echo '$USER'")
            self.ssh.username = self._remote_user
        return self._remote_user

    @property
    def remote_home(self):
        """
        Get (or create) a remote home
        """
        if not self._remote_home:
            self._remote_home = self.ssh.execute_or_fail(
                "echo '%s'" % self.ssh.settings.tunel_remote_home, quiet=True
            )
            self.ssh.execute_or_fail("mkdir -p %s" % self.remote_assets_dir)
        return self._remote_home

    @property
    def remote_work(self):
        """
        Get a remote work (must exist)
        """
        if not self._remote_work:
            self._remote_work = self.ssh.execute_or_fail(
                "echo '%s'" % self.ssh.settings.tunel_remote_work
            )
        return self._remote_work

    @property
    def remote_sockets(self):
        """
        Get a remote sockets directory
        """
        if self.ssh.settings.tunel_remote_sockets:
            return self.ssh.settings.tunel_remote_sockets
        return os.path.join(self.remote_home, ".tunel", "sockets")

    def scp_get(self, src, dest):
        """
        Copy an asset FROM the server to local assets
        """
        dirname = os.path.dirname(dest)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        return self.ssh.scp_from(src, dest)

    def scp_and_run(self, script):
        """
        Given a local script, copy to cluster, run, and return the result
        """
        # Remote modules file we will read
        remote_script = os.path.join(self.remote_assets_dir, script)

        # Local script to generate it (doesn't work interactivelt)
        local_script = os.path.join(self.assets, script)

        # Copy to the remote
        self.ssh.scp_to(local_script, remote_script)

        return self.ssh.execute(
            "chmod u+x %s; /bin/bash -l %s" % (remote_script, remote_script)
        )



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"


from tunel.ssh import Tunnel


def main(args, parser, extra, subparser):

    tunel = Tunnel(
        args.server,
        remote_port=args.port,
        local_port=args.local_port,
        isolated_nodes=args.isolated_nodes,
    )
    tunel.tunnel()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

import inspect
import logging as _logging
import os
import platform
import sys
import threading

from rich import print
from rich.console import Console, Group
from rich.panel import Panel
from rich.rule import Rule
from rich.syntax import Syntax
from rich.table import Table


class ColorizingStreamHandler(_logging.StreamHandler):

    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[%dm"
    BOLD_SEQ = "\033[1m"

    colors = {
        "WARNING": YELLOW,
        "INFO": GREEN,
        "DEBUG": BLUE,
        "CRITICAL": RED,
        "ERROR": RED,
    }

    def __init__(self, nocolor=False, stream=sys.stderr, use_threads=False):
        super().__init__(stream=stream)
        self._output_lock = threading.Lock()
        self.nocolor = nocolor or not self.can_color_tty()

    def can_color_tty(self):
        if "TERM" in os.environ and os.environ["TERM"] == "dumb":
            return False
        return self.is_tty and not platform.system() == "Windows"

    @property
    def is_tty(self):
        isatty = getattr(self.stream, "isatty", None)
        return isatty and isatty()

    def emit(self, record):
        with self._output_lock:
            try:
                self.format(record)  # add the message to the record
                self.stream.write(self.decorate(record))
                self.stream.write(getattr(self, "terminator", "\n"))
                self.flush()
            except BrokenPipeError as e:
                raise e
            except (KeyboardInterrupt, SystemExit):
                # ignore any exceptions in these cases as any relevant messages have been printed before
                pass
            except Exception:
                self.handleError(record)

    def decorate(self, record):
        message = record.message
        message = [message]
        if not self.nocolor and record.levelname in self.colors:
            message.insert(0, self.COLOR_SEQ % (30 + self.colors[record.levelname]))
            message.append(self.RESET_SEQ)
        return "".join(message)


class Logger:
    def __init__(self):
        self.logger = _logging.getLogger(__name__)
        self.log_handler = [self.text_handler]
        self.stream_handler = None
        self.printshellcmds = False
        self.quiet = False
        self.logfile = None
        self.last_msg_was_job_info = False
        self.logfile_handler = None
        self.c = Console()

    def purple(self, msg):
        self.c.print("[bold purple]%s" % msg)

    def cyan(self, msg):
        self.c.print("[bold cyan]%s" % msg)

    def panel_group(self, msgs):
        """
        Input is a dictionary of colors (for panel background) and corresponding messages.
        """
        panels = [Panel(v, style="on %s" % k) for k, v in msgs.items()]
        panel_group = Group(*panels)
        print(Panel(panel_group))

    def cleanup(self):
        if self.logfile_handler is not None:
            self.logger.removeHandler(self.logfile_handler)
            self.logfile_handler.close()
        self.log_handler = [self.text_handler]

    def handler(self, msg):
        for handler in self.log_handler:
            handler(msg)

    def set_stream_handler(self, stream_handler):
        if self.stream_handler is not None:
            self.logger.removeHandler(self.stream_handler)
        self.stream_handler = stream_handler
        self.logger.addHandler(stream_handler)

    def set_level(self, level):
        self.logger.setLevel(level)

    def location(self, msg):
        callerframerecord = inspect.stack()[1]
        frame = callerframerecord[0]
        info = inspect.getframeinfo(frame)
        self.debug(
            "{}: {info.filename}, {info.function}, {info.lineno}".format(msg, info=info)
        )

    def info(self, msg):
        self.handler(dict(level="info", msg=msg))

    def warning(self, msg):
        self.handler(dict(level="warning", msg=msg))

    def debug(self, msg):
        self.handler(dict(level="debug", msg=msg))

    def error(self, msg):
        self.handler(dict(level="error", msg=msg))

    def exit(self, msg, return_code=1):
        self.handler(dict(level="error", msg=msg))
        sys.exit(return_code)

    def progress(self, done=None, total=None):
        self.handler(dict(level="progress", done=done, total=total))

    def shellcmd(self, msg):
        if msg is not None:
            msg = dict(level="shellcmd", msg=msg)
            self.handler(msg)

    def print_pretty(self, obj):
        """
        Print a dict (and nested content) pretty.
        """
        for field, content in obj.items():

            self.c.print(Rule(field, style="blue"))
            # If we have a dict, make key value pairs
            if isinstance(content, dict):
                panel = ""
                for i, k in enumerate(content):
                    v = content[k]
                    panel += f"{k}: {v}"
                    if i != len(content) - 1:
                        panel += "\n"
                self.c.print(Panel(panel))

            # If we have a dict, make key value pairs
            elif isinstance(content, list) and isinstance(content[0], str):
                panel = ""
                for i, item in enumerate(content):
                    panel += f"- {item}"
                    if i != len(content) - 1:
                        panel += "\n"
                self.c.print(Panel(panel))

            # List of dicts, assumes same keys
            elif isinstance(content, list) and content and isinstance(content[0], dict):
                table = Table()
                fields = content[0]
                for i, key in enumerate(fields):
                    color = "magenta"
                    if i == 0:
                        color = "cyan"
                    table.add_column(key, style=color)
                count = 0
                for item in content:
                    row = []
                    for k in fields:
                        if k in item:
                            row.append(item[k])
                    table.add_row(*row)
                    count += 1
                self.c.print(table)
            elif field == "examples":
                self.c.print(Panel(Syntax(content, "bash")))
            else:
                self.c.print(Panel(content))

    def text_handler(self, msg):
        """
        The default handler prints output to the console.
        Args:
            msg (dict):     the log message dictionary
        """
        level = msg["level"]
        if level == "info" and not self.quiet:
            self.logger.info(msg["msg"])
        if level == "warning":
            self.logger.warning(msg["msg"])
        elif level == "error":
            self.logger.error(msg["msg"])
        elif level == "debug":
            self.logger.debug(msg["msg"])
        elif level == "progress" and not self.quiet:
            done = msg["done"]
            total = msg["total"]
            p = done / total
            percent_fmt = ("{:.2%}" if p < 0.01 else "{:.0%}").format(p)
            self.logger.info(
                "{} of {} steps ({}) done".format(done, total, percent_fmt)
            )
        elif level == "shellcmd":
            if self.printshellcmds:
                self.logger.warning(msg["msg"])


logger = Logger()


def setup_logger(
    quiet=False,
    printshellcmds=False,
    nocolor=False,
    stdout=False,
    debug=False,
    use_threads=False,
    wms_monitor=None,
):
    # console output only if no custom logger was specified
    stream_handler = ColorizingStreamHandler(
        nocolor=nocolor,
        stream=sys.stdout if stdout else sys.stderr,
        use_threads=use_threads,
    )
    logger.set_stream_handler(stream_handler)
    logger.set_level(_logging.DEBUG if debug else _logging.INFO)
    logger.quiet = quiet
    logger.printshellcmds = printshellcmds



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import shlex
import threading
import time

import tunel.utils as utils
from tunel.launcher.base import Launcher
from tunel.logger import logger


class Docker(Launcher):
    name = "docker"

    def run(self, *args, **kwargs):
        """
        Run handles some command to docker (this can be expanded in functionality)
        """
        if self.path:
            command = "docker run -d --rm -i %s PATH=%s %s" % (
                self.env_command,
                self.path,
                " ".join(args[0]),
            )
        else:
            command = "docker run -d --rm -i %s %s" % (
                self.env_command,
                " ".join(args[0]),
            )

        res = self.ssh.execute(command)
        self.ssh.print_output(res)

    @property
    def env_command(self):
        """
        Transform a listing of key=value into docker --env flags
        """
        envars = self.environ.split(" ") if self.environ else ""
        if envars:
            envars = ["--env %s" for e in envars]
            envars = " ".join(envars)
        return envars

    def run_app(self, app):
        """
        Run a docker container directly on your remote resource.
        """
        # Make sure we set the username to the ssh
        self.username

        # Add any paths from the config
        paths = self.settings.get("paths", [])

        # Prepare dictionary with content to render into recipes
        render = self.prepare_render(app, paths)
        render["docker"] = self.name

        # Clean up previous sockets
        self.ssh.execute(["rm", "-rf", "%s/*.sock" % render["scriptdir"]])

        # Load the app template
        template = app.load_template()
        result = template.render(**render)

        # Write script to temporary file
        tmpfile = utils.get_tmpfile()
        utils.write_file(tmpfile, result)

        # Copy over to server
        remote_script = os.path.join(self.remote_assets_dir, app.name, app.script)
        self.ssh.scp_to(tmpfile, remote_script)

        # Instead of a Singularity command, we run the script
        logger.c.print()
        logger.c.print("== INSTRUCTIONS WILL BE PRINTED with a delay ==")
        command = "%s %s %s %s" % (
            self.path,
            self.environ,
            self.ssh.settings.shell,
            remote_script,
        )
        self.print_tunnel_instructions(app, render["socket"])
        self.ssh.execute(command, stream=True)

    def stop_app(self, app):
        """
        Wrapper to stop a single app.
        """
        return self.stop([app.name])

    def stop(self, names):
        """
        Stop one or more named jobs
        """
        for name in names:
            name = name.replace("/", "-")
            logger.purple("Killing %s docker container on %s" % (name, self.ssh.server))
            self.ssh.execute_or_fail("docker stop %s" % name)
            self.ssh.execute_or_fail("docker rm %s || echo Already removed." % name)

    def print_tunnel_instructions(self, app, socket):
        """
        Start a separate thread to print connection details.
        """
        thread = threading.Thread(
            target=post_commands,
            name="Logger",
            args=[self.ssh, app, socket],
        )
        thread.start()


def post_commands(ssh, app, socket):
    time.sleep(10)
    ssh.tunnel_login_node(socket=socket, app=app)
    if app.post_command:
        logger.info("Found post command %s" % app.post_command)
        post = app.post_command.replace("$socket_dir", os.path.dirname(socket))
        ssh.execute(shlex.split(post), stream=True)
    time.sleep(30)
    ssh.tunnel_login_node(socket=socket, app=app)
